//---------------------------------------------------------------------------
#include "http_utils.h"
//---------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//---------------------------------------------------------------------------
int str_length(char *s)
{
    char *p = s;
    while (*p)
        p++;
    return p - s;
}
//---------------------------------------------------------------------------
char *find_str(char *beg, char *end, char *s)
{
    char *res = 0;

    int s_length = str_length(s);

    if (end - beg >= s_length)
    {
        end -= s_length - 1;

        while (beg != end)
        {
            char *p = beg;
            char *t = s;

            bool good = true;

            while (*t)
            {
                if (*p != *t)
                {
                    good = false;
                    break;
                }
                p++;
                t++;
            }

            if (good)
            {
                res = beg;
                break;
            }

            beg++;
        }
    }

    return res;
}
//---------------------------------------------------------------------------
//method uri protocol_version
//name: value
//name: value
//
bool parse_header
(
    char *beg, char *end, char **pos,
    std::string * const method,
    char **uri_beg, char **uri_end,
    std::string * const protocol_version,
    std::map<std::string, std::string> * const params
)
{
    bool res = true;

    if (method) method->clear();
    if (uri_beg) *uri_beg = 0;
    if (uri_end) *uri_end = 0;
    if (protocol_version) protocol_version->clear();
    if (params) params->clear();

    while (true)
    {
        char *p = find_str(beg, end, " ");
        if (!p)
        {
            res = false;
            break;
        }
        if (method) method->append(beg, p - beg);
        beg = p + 1;

        p = find_str(beg, end, " ");
        if (!p)
        {
            res = false;
            break;
        }
        if (uri_beg) *uri_beg = beg;
        if (uri_end) *uri_end = p;
        beg = p + 1;

        p = find_str(beg, end, "\r\n");
        if (!p)
        {
            res = false;
            break;
        }
        if (protocol_version) protocol_version->append(beg, p - beg);
        beg = p + 2;

        while (true)
        {
            if (end - beg >= 2 && *beg == '\r' && *(beg + 1) == '\n')
            {
                beg += 2;
                break;
            }

            p = find_str(beg, end, ": ");
            if (!p)
            {
                res = false;
                break;
            }
            std::string name(beg, p - beg);
            beg = p + 2;

            p = find_str(beg, end, "\r\n");
            if (!p)
            {
                res = false;
                break;
            }
            std::string value(beg, p - beg);
            beg = p + 2;

            if (params) params->insert(std::map<std::string, std::string>::value_type(name, value));
        }

        break;
    }

    if (pos) *pos = beg;

    return res;
}
//---------------------------------------------------------------------------
//protocol://login:password@host:port/path?query#fragment
bool parse_uri
(
    char *beg, char *end, char **pos,
    std::string * const protocol,
    std::string * const login,
    std::string * const password,
    std::string * const host,
    int * const port,
    std::string * const path,
    char **query_beg, char **query_end,
    std::string * const fragment
)
{
    bool res = true;

    if (protocol) protocol->clear();
    if (login) login->clear();
    if (password) password->clear();
    if (host) host->clear();
    if (port) *port = 0;
    if (path) path->clear();
    if (query_beg) *query_beg = 0;
    if (query_end) *query_end = 0;
    if (fragment) fragment->clear();

    while (true)
    {
        char *p = find_str(beg, end, "://");
        if (p)
        {
            if (protocol) protocol->append(beg, p - beg);
            beg = p + 3;
        }

        p = find_str(beg, end, "@");
        if (p)
        {
            char *t = find_str(beg, p, ":");
            if (!t)
            {
                res = false;
                break;
            }
            if (login) login->append(beg, t - beg);
            if (password) password->append(t + 1, p - t - 1);
            beg = p + 1;
        }

        p = find_str(beg, end, "/");
        if (p)
        {
            char *t = find_str(beg, p, ":");
            if (t)
            {
                if (host) host->append(beg, t - beg);
                if (port) *port = atoi(t + 1);
            }
            else
            {
                if (host) host->append(beg, p - beg);
            }
            beg = p + 1;
        }

        p = find_str(beg, end, "?");
        if (p)
        {
            if (path) path->append(beg, p - beg);
            beg = p + 1;

            p = find_str(beg, end, "#");
            if (p)
            {
                if (query_beg) *query_beg = beg;
                if (query_end) *query_end = p;
                if (fragment) fragment->append(p + 1, end - p - 1);
            }
            else
            {
                if (query_beg) *query_beg = beg;
                if (query_end) *query_end = end;
            }
        }
        else
        {
            p = find_str(beg, end, "#");
            if (p)
            {
                if (path) path->append(beg, p - beg);
                if (fragment) fragment->append(p + 1, end - p - 1);
            }
            else
            {
                if (path) path->append(beg, end - beg);
            }
        }
        beg = end;

        break;
    }

    if (pos) *pos = beg;

    return res;
}
//---------------------------------------------------------------------------
//protocol://login:password@host_port/path?query#fragment
std::string create_uri
(
    const std::string &protocol,
    const std::string &login,
    const std::string &password,
    const std::string &host_port,
    const std::string &path,
    const std::string &query,
    const std::string &fragment
)
{
    std::string res;
    if (protocol.length()) res.append(protocol + "://");
    if (login.length()) res.append(login + ":" + password + "@");
    res.append(host_port + "/");
    if (path.length()) res.append(path);
    if (query.length()) res.append("?" + query);
    if (fragment.length()) res.append("#" + fragment);
    return res;
}
//---------------------------------------------------------------------------
std::string get_content_type(const std::string &ext)
{
    return
    ext == "htm"? "text/html" :
    ext == "html"? "text/html" :
    ext == "ico"? "image/vnd.microsoft.icon" :
    ext == "js"? "application/javascript" :
    ext == "jpg"?"image/jpeg" :
    ext == "swf"? "application/x-shockwave-flash" :
    ext == "css"? "text/css" :
    "application/octet-stream";
}
//---------------------------------------------------------------------------
//name=value&name=value
bool parse_query
(
    char *beg, char *end, char **pos,
    std::map<std::string, std::string> * const params
)
{
    bool res = true;

    params->clear();

    while (true)
    {
        if (beg == end) break;

        while (true)
        {
            char *p = find_str(beg, end, "=");
            if (!p)
            {
                res = false;
                break;
            }
            std::string name(beg, p - beg);
            beg = p + 1;

            std::string value;
            p = find_str(beg, end, "&");
            if (!p)
            {
                value.append(beg, end - beg);
                beg = end;

                params->insert(std::map<std::string, std::string>::value_type(name, value));
                break;
            }
            value.append(beg, p - beg);
            beg = p + 1;

            params->insert(std::map<std::string, std::string>::value_type(name, value));
        }

        break;
    }

    if (pos) *pos = beg;

    return res;
}
//---------------------------------------------------------------------------
std::string http_time_str(const std::time_t t)
{
    static const char *days = "Sun\0Mon\0Tue\0Wed\0Thu\0Fri\0Sat";
    static const char *months = "Jan\0Feb\0Mar\0Apr\0May\0Jun\0Jul\0Aug\0Sep\0Oct\0Nov\0Dec";

    static char buffer[30];

    std::tm *_t = std::gmtime(&t);

    sprintf
    (
        buffer,
        "%s, %02u %s %04u %02u:%02u:%02u GMT",
        days + _t->tm_wday * 4, (unsigned)_t->tm_mday, months + _t->tm_mon * 4, (unsigned)_t->tm_year + 1900,
        (unsigned)_t->tm_hour, (unsigned)_t->tm_min, (unsigned)_t->tm_sec
    );

    return buffer;
}
//---------------------------------------------------------------------------
std::string get_ext(const std::string &s)
{
    int p_pos = s.find_last_of('.');
    int s_pos = s.find_last_of("\\/");
    return
    p_pos == std::string::npos ||
    (s_pos != std::string::npos && s_pos > p_pos)? std::string() : s.substr(p_pos + 1);
}
//---------------------------------------------------------------------------
std::string get_name(const std::string &s)
{
    int pos = s.find_last_of("\\/");
    return pos == std::string::npos? s : s.substr(pos + 1);
}
//---------------------------------------------------------------------------
std::string change_ext(const std::string &name, const std::string &ext)
{
    int p_pos = name.find_last_of('.');
    int s_pos = name.find_last_of("\\/");
    return
    p_pos == std::string::npos ||
    (s_pos != std::string::npos && s_pos > p_pos)? name + ext : name.substr(0, p_pos) + ext;
}
//---------------------------------------------------------------------------
std::string get_path(const std::string &s)
{
    int pos = s.find_last_of("\\/");
    return pos == std::string::npos? std::string() : s.substr(0, pos + 1);
}
//---------------------------------------------------------------------------
int get_level_min(const std::string &path)
{
    int res = 0;
    int level = 0;
    int pos = 0;
    while (true)
    {
        int slash = path.find_first_of("\\/", pos);
        int dots = path.find("..", pos);

        if (slash == std::string::npos)
        {
            if (dots == std::string::npos) break;
            pos = dots + 2;
            level--;
            if (level < res) res = level;
        }
        else if (dots == std::string::npos)
        {
            if (slash == std::string::npos) break;
            pos = slash + 1;
            level++;
        }
        else
        {
            if (slash > dots)
            {
                pos = dots + 2;
                level--;
                if (level < res) res = level;
            }
            else
            {
                pos = slash + 1;
                level++;
            }
        }
    }
    return res;
}
//---------------------------------------------------------------------------
int http_print
(
    char *buffer,
    int * const sizeof_header,
    const char *code_and_reason,
    const char *lines,
    const char *content_type,
    const char *content
)
{
    int _sizeof_header =
    sprintf
    (
        buffer,
        "HTTP/1.1 %s\r\n"
        "%s"
        "Content-Type: %s\r\n"
        "Content-Length: %i\r\n"
        "\r\n",
        code_and_reason,
        lines? lines : "",
        content_type? content_type : "text/plain",
        (int)strlen(content? content : code_and_reason)
    );
    if (sizeof_header) *sizeof_header = _sizeof_header;
    return _sizeof_header + sprintf
    (
        buffer + _sizeof_header,
        "%s",
        content? content : code_and_reason
    );
}
//---------------------------------------------------------------------------
//name=value
//name=value
//
bool parse_params
(
    char *beg, char *end,
    std::map<std::string, std::string> * const params
)
{
    bool res = true;

    params->clear();

    while (true)
    {
        if (end - beg == 2 && *beg == '\r' && *(beg + 1) == '\n')
        {
            beg += 2;
            break;
        }

        char *p = find_str(beg, end, "=");
        if (!p)
        {
            res = false;
            break;
        }
        std::string name(beg, p - beg);
        beg = p + 1;

        p = find_str(beg, end, "\r\n");
        if (!p)
        {
            res = false;
            break;
        }
        std::string value(beg, p - beg);
        beg = p + 2;

        if (params) params->insert(std::map<std::string, std::string>::value_type(name, value));
    }

    return res;
}
