//------------------------------------------------------------------------------
//Copyright (C) 2013 by Gusev Alexander
//e-mail: alphacity@rambler.ru
//---------------------------------------------------------------------------
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
//---------------------------------------------------------------------------
#include "platform.h"
#include "socket_engine.h"
#include "printf_ex.h"
#include "time_utils.h"
#include "proxy_utils.h"
#include "http_utils.h"
#define DEBUG
//---------------------------------------------------------------------------
static std::string login = "user";
static std::string password = "0000";
static std::string root;
static int sizeof_in = 10240;
static int sizeof_out = 1048576;
//---------------------------------------------------------------------------
static std::string exe_name;
//---------------------------------------------------------------------------
class simple_server_node:
    public se_server_node
{
public:
    bool etx; //  
    bool authorized;
    int a_counter;
    static int max_a_counter;

    int sizeof_in;
    char *in;
    int sizeof_out;
    char *out;

    int q;

    simple_server_node(void):
        authorized(false),
        a_counter(0),
        sizeof_in(::sizeof_in),
        in(new char[sizeof_in]),
        sizeof_out(::sizeof_out),
        out(new char[sizeof_out]),
        q(50)
    {
        sizeof_buffer = sizeof_in;
        buffer = in;
    }
    virtual ~simple_server_node(void)
    {
        disconnect();

        delete [] in;
        delete [] out;
    }

    virtual void before_job_proc(void)
    {
        if (job_active() && get_job() == _recv && sizeof_data_cur >= 4)
        {
            //     
            //  -   sizeof_in
            //    ( - "\r\n\r\n")
            //POST-   (   "\r\n\r\n")
            
            etx =
            in[sizeof_data_cur - 4] == '\r' &&
            in[sizeof_data_cur - 3] == '\n' &&
            in[sizeof_data_cur - 2] == '\r' &&
            in[sizeof_data_cur - 1] == '\n';

            if (etx)
            {
                #ifdef DEBUG
                in[sizeof_data_cur - 2] = '\0';
                printf_ex
                (
                    "//---------------------------------------------------------------------------\n"
                    "%s node (ip = %s, port = %hu) etx detected\n"
                    "//---------------------------------------------------------------------------\n"
                    "%s\n"
                    "//---------------------------------------------------------------------------\n",
                    system_time_str().c_str(),
                    get_ip().c_str(),
                    get_port(),
                    in
                );
                in[sizeof_data_cur - 2] = '\r';
                #endif

                abort_job();

                char *in = this->in;
                char *in_end = in + sizeof_data_cur;

                char *out = this->out;
                int sizeof_out = this->sizeof_out;
                int sizeof_data;
                this->sizeof_data = 0;

                while (true)
                {
                    char *pos;
                    std::string method;
                    char *uri_beg, *uri_end;
                    std::string protocol_version;
                    std::map<std::string, std::string> params;

                    if
                    (
                        !parse_header(in, in_end, &pos, &method, &uri_beg, &uri_end, &protocol_version, &params) ||
                        protocol_version != "HTTP/1.1"
                    )
                    {
                        disconnect();
                        return;
                    }

                    std::string protocol;
                    std::string login;
                    std::string password;
                    std::string host;
                    int port;
                    std::string path;
                    char *query_beg;
                    char *query_end;
                    std::string fragment;

                    if
                    (
                        !parse_uri(uri_beg, uri_end, 0, &protocol, &login, &password, &host, &port, &path, &query_beg, &query_end, &fragment) ||
                        (protocol.length() && protocol != "http")
                    )
                    {
                        disconnect();
                        return;
                    }

                    std::string host_port;
                    std::map<std::string, std::string>::iterator i = params.find("Host");
                    if (i != params.end()) host_port = i->second;

                    int sizeof_header;

                    if (!authorized)
                    {
                        if (a_counter > max_a_counter)
                        {
                            sizeof_data = http_print(out, &sizeof_header, "403 Forbidden");
                        }
                        else
                        {
                            a_counter++;

                            if (login == ::login && password == ::password)
                            {
                                authorized = true;

                                #ifdef DEBUG
                                printf_ex
                                (
                                    "%s node (ip = %s, port = %hu) authorization detected\n",
                                    system_time_str().c_str(),
                                    get_ip().c_str(),
                                    get_port()
                                );
                                #endif
                            }
                            else
                            {
                                std::string authorization;
                                std::map<std::string, std::string>::iterator i = params.find("Authorization");
                                if (i != params.end()) authorization = i->second;

                                if (authorization == "Basic " + base64_encode((const unsigned char *)(::login + ":" + ::password).c_str(), ::login.length() + 1 + ::password.length()))
                                {
                                    authorized = true;

                                    #ifdef DEBUG
                                    printf_ex
                                    (
                                        "%s node (ip = %s, port = %hu) authorization detected\n",
                                        system_time_str().c_str(),
                                        get_ip().c_str(),
                                        get_port()
                                    );
                                    #endif
                                }
                                else
                                {
                                    sizeof_data = http_print(out, &sizeof_header, "401 Unauthorized", ("WWW-Authenticate: Basic realm=\"" + change_ext(get_name(exe_name), "") + "\"\r\n").c_str());
                                }
                            }
                        }
                    }

                    if (authorized)
                    {
                        if (method == "GET")
                        {
                            if (!parse_query(query_beg, query_end, 0, &params))
                            {
                                disconnect();
                                return;
                            }

                            std::string root_path = root + path;

                            if (root_path.at(root_path.length() - 1) == '\\' || root_path.at(root_path.length() - 1) == '/') root_path += "index.htm";

                            if (get_level_min(path) < 0)
                            {
                                sizeof_data = http_print(out, &sizeof_header, "403 Forbidden");
                            }
                            else
                            {
                                struct stat si;
                                if (stat(root_path.c_str(), &si))
                                {
                                    sizeof_data = http_print(out, &sizeof_header, "404 Not Found");
                                }
                                else
                                {
                                    if (S_ISDIR(si.st_mode))
                                    {
                                        sizeof_data = http_print(out, &sizeof_header, "302 Found", ("Location: " + create_uri("http", login, password, host_port, path + "/", std::string(query_beg, query_end - query_beg), fragment) + "\r\n").c_str());
                                    }
                                    else
                                    {
                                        std::string ext = get_ext(root_path);

                                        sizeof_header =
                                        sprintf
                                        (
                                            out,
                                            "HTTP/1.1 200 OK\r\n"
                                            "Last-Modified: %s\r\n"
                                            "Content-Type: %s\r\n"
                                            "Content-Length: %li\r\n"
                                            "\r\n",
                                            http_time_str(si.st_mtime).c_str(),
                                            get_content_type(ext).c_str(),
                                            (long)si.st_size
                                        );

                                        sizeof_data = sizeof_header + si.st_size;

                                        if (sizeof_data > sizeof_out)
                                        {
                                            sizeof_data = http_print(out, &sizeof_header, "500 Internal Server Error");
                                        }
                                        else
                                        {
                                            FILE *f = fopen(root_path.c_str(), "rb");
                                            if (!f)
                                            {
                                                sizeof_data = http_print(out, &sizeof_header, "500 Internal Server Error");
                                            }
                                            else
                                            {
                                                fread(out + sizeof_header, si.st_size, 1, f);
                                                fclose(f);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else
                        {
                            disconnect();
                            return;
                        }
                    }

                    if (out[13] != 'O' || out[14] != 'K')
                    {
                        #ifdef DEBUG
                        printf_ex
                        (
                            "%s node (ip = %s, port = %hu) error detected\n",
                            system_time_str().c_str(),
                            get_ip().c_str(),
                            get_port()
                        );
                        #endif
                    }

                    #ifdef DEBUG
                    out[sizeof_header - 2] = '\0';
                    printf_ex
                    (
                        "//---------------------------------------------------------------------------\n"
                        "%s node (ip = %s, port = %hu) header to send\n"
                        "//---------------------------------------------------------------------------\n"
                        "%s\n"
                        "//---------------------------------------------------------------------------\n",
                        system_time_str().c_str(),
                        get_ip().c_str(),
                        get_port(),
                        out
                    );
                    out[sizeof_header - 2] = '\r';
                    #endif

                    out += sizeof_data;
                    sizeof_out -= sizeof_data;
                    this->sizeof_data += sizeof_data;

                    in = pos;
                    if (in == in_end) break;
                }

                sizeof_buffer = this->sizeof_out;
                buffer = this->out;

                if (activate_job(_send))
                {
                    #ifdef DEBUG
                    printf_ex
                    (
                        "%s node (ip = %s, port = %hu) job %s activated\n",
                        system_time_str().c_str(),
                        get_ip().c_str(),
                        get_port(),
                        get_job_str(get_job())
                    );
                    #endif
                }
                else
                {
                    #ifdef DEBUG
                    printf_ex
                    (
                        "%s node (ip = %s, port = %hu) can not activate job %s\n",
                        system_time_str().c_str(),
                        get_ip().c_str(),
                        get_port(),
                        get_job_str(get_job())
                    );
                    #endif
                }
            }
        }
    }
    virtual void result_proc(void)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s node (ip = %s, port = %hu) job %s finished with result %s\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port(),
            get_job_str(get_job()),
            get_result_str(get_result())
        );
        #endif

        if (get_result() == aborted)
        {
            return;
        }

        if (get_result() != success)
        {
            if (get_result() != connection_closed) disconnect();
            return;
        }

        if (get_job() == _send)
        {
            sizeof_buffer = sizeof_in;
            buffer = in;
            sizeof_data = sizeof_buffer;

            if (activate_job(_recv, true))
            {
                #ifdef DEBUG
                printf_ex
                (
                    "%s node (ip = %s, port = %hu) job %s activated\n",
                    system_time_str().c_str(),
                    get_ip().c_str(),
                    get_port(),
                    get_job_str(get_job())
                );
                #endif
            }
            else
            {
                #ifdef DEBUG
                printf_ex
                (
                    "%s node (ip = %s, port = %hu) can not activate job %s\n",
                    system_time_str().c_str(),
                    get_ip().c_str(),
                    get_port(),
                    get_job_str(get_job())
                );
                #endif
            }
        }
    }
    virtual void after_connect(void)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s node (ip = %s, port = %hu) connected\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port()
        );
        #endif

        sizeof_buffer = sizeof_in;
        buffer = in;
        sizeof_data = sizeof_buffer;

        if (activate_job(_recv, true))
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s node (ip = %s, port = %hu) job %s activated\n",
                system_time_str().c_str(),
                get_ip().c_str(),
                get_port(),
                get_job_str(get_job())
            );
            #endif
        }
        else
        {
            #ifdef DEBUG
            printf_ex
            (
                "%s node (ip = %s, port = %hu) can not activate job %s\n",
                system_time_str().c_str(),
                get_ip().c_str(),
                get_port(),
                get_job_str(get_job())
            );
            #endif
        }
    }
    virtual void after_disconnect(void)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s node (ip = %s, port = %hu) disconnected\n",
            system_time_str().c_str(),
            get_ip().c_str(),
            get_port()
        );
        #endif
    }
};

int simple_server_node::max_a_counter = 3;

class simple_se_server:
    public se_server
{
public:
    virtual void after_error(const errors error)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s server function call finished with error %s\n",
            system_time_str().c_str(),
            get_error_str(error)
        );
        #endif
    }

    virtual simple_server_node *new_node(void)
    {
        return new simple_server_node();
    }
};
//---------------------------------------------------------------------------
static simple_se_server server;
//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
    exe_name = argv[0];

    #ifdef DEBUG
    printf_ex_init(true, true, change_ext(exe_name, ".log").c_str(), "w");
    printf_ex
    (
        "//---------------------------------------------------------------------------\n"
        "%s logging started\n",
        system_time_str().c_str()
    );
    #endif

    std::string ip = "0";
    unsigned short port = 80;
    root = (get_path(exe_name) + "root/").c_str();

    std::ifstream is(change_ext(exe_name, ".cfg").c_str());
    std::string params_str;
    if (!is)
    {
        printf_ex("////can not read cfg-file\n");
    }
    else
    {
        while (is.peek() != EOF)
        {
            std::string line;
            std::getline(is, line);
            if (line.length() && line.at(line.length() - 1) == '\r') line.resize(line.length() - 1);
            if (line.length() && (line.length() < 2 || line.at(0) != '/' || line.at(1) != '/'))
            {
                params_str.append(line);
                params_str.append("\r\n");
            }
        }
        params_str.append("\r\n");
        is.close();
    }

    char *params_s = (char *)params_str.c_str();
    std::map<std::string, std::string> params;
    std::map<std::string, std::string>::iterator i;
    if (!parse_params(params_s, params_s + params_str.length(), &params))
    {
        printf_ex("////can not parse params\n");
    }
    else
    {
        i = params.find("ip");
        if (i != params.end()) ip = i->second;

        i = params.find("port");
        if (i != params.end()) port = atoi(i->second.c_str());

        i = params.find("login");
        if (i != params.end()) login = i->second;

        i = params.find("password");
        if (i != params.end()) password = i->second;

        i = params.find("root");
        if (i != params.end()) root = i->second;

        i = params.find("sizeof_in");
        if (i != params.end()) sizeof_in = atoi(i->second.c_str());

        i = params.find("sizeof_out");
        if (i != params.end()) sizeof_out = atoi(i->second.c_str());
    }

    printf_ex("////params:\n");
    printf_ex("////ip=%s\n", ip.c_str());
    printf_ex("////port=%hi\n", port);
    printf_ex("////login=%s\n", login.c_str());
    printf_ex("////password=%s\n", password.c_str());
    printf_ex("////root=%s\n", root.c_str());
    printf_ex("////sizeof_in=%i\n", sizeof_in);
    printf_ex("////sizeof_out=%i\n", sizeof_out);

    #ifndef IS_LINUX

    WSADATA wsa_data;
    if (WSAStartup(0x101, &wsa_data) || wsa_data.wVersion != 0x101)
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s can not initialize sockets\n",
            system_time_str().c_str()
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s sockets initialized\n",
            system_time_str().c_str()
        );
        #endif
    }

    #endif

    server.prepare
    (
        AF_INET,
        SOCK_STREAM,
        IPPROTO_TCP,
        ip.c_str(),
        port
    );
    if (!server.activate(true))
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s can not activate server\n",
            system_time_str().c_str()
        );
        #endif
    }
    else
    {
        #ifdef DEBUG
        printf_ex
        (
            "%s server activated\n",
            system_time_str().c_str()
        );
        #endif
    }

    while (true)
    {
        server.job_proc();
        Sleep(100);
    }

    server.deactivate();

    #ifndef IS_LINUX

    WSACleanup();

    #endif

    #ifdef DEBUG
    printf_ex
    (
        "%s sockets uninitialized\n",
        system_time_str().c_str()
    );
    #endif

    printf_ex_uninit();

    return 0;
}
//---------------------------------------------------------------------------

