//------------------------------------------------------------------------------
#include "printf_ex.h"
//------------------------------------------------------------------------------
#include <stdio.h>
#include <stdarg.h>
//------------------------------------------------------------------------------
#include "mt_obj.h"
//------------------------------------------------------------------------------
static bool print_to_file;
static bool print_to_screen_also;
static FILE *f;
static bool enabled = true;
static mt_obj cs;
//------------------------------------------------------------------------------
bool printf_ex_init
(
    const bool print_to_file,
    const bool print_to_screen_also,
    const char *file_name, const char *file_mode
)
{
    bool res = false;

    ::print_to_file = print_to_file;
    ::print_to_screen_also = print_to_screen_also;

    if (!print_to_file)
    {
        res = true;
    }
    else
    {
        if (file_name && file_mode)
        {
            f = fopen(file_name, file_mode);
            res = f;
        }
    }

    return res;
}
//------------------------------------------------------------------------------
void printf_ex_uninit(void)
{
    fclose(f);
    f = 0;
}
//------------------------------------------------------------------------------
void printf_ex_enabled(const bool enabled)
{
    ::enabled = enabled;
}
//------------------------------------------------------------------------------
bool printf_ex(const char *format, ...)
{
    bool res = false;

    if (!enabled)
    {
        res = true;
    }
    else
    {
        cs.lock();

        if (!print_to_file)
        {
            va_list arglist;

            va_start(arglist, format);
            res = vprintf(format, arglist) != EOF;
            va_end(arglist);
        }
        else
        {
            if (print_to_screen_also)
            {
                va_list arglist;

                va_start(arglist, format);
                res = vprintf(format, arglist) != EOF;
                va_end(arglist);
            }

            if (f)
            {
                va_list arglist;

                va_start(arglist, format);
                bool _res = vfprintf(f, format, arglist) != EOF && !fflush(f);
                res = print_to_screen_also? res && _res : _res;
                va_end(arglist);
            }
        }

        cs.unlock();
    }

    return res;
}

