//------------------------------------------------------------------------------
#include "proxy_utils.h"
//------------------------------------------------------------------------------
#include <sstream>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
//------------------------------------------------------------------------------
static const std::string base64_chars =
             "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
             "abcdefghijklmnopqrstuvwxyz"
             "0123456789+/";

std::string base64_encode(unsigned char const* bytes_to_encode, unsigned int in_len) {
  std::string ret;
  int i = 0;
  int j = 0;
  unsigned char char_array_3[3];
  unsigned char char_array_4[4];

  while (in_len--) {
    char_array_3[i++] = *(bytes_to_encode++);
    if (i == 3) {
      char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
      char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
      char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
      char_array_4[3] = char_array_3[2] & 0x3f;

      for(i = 0; (i <4) ; i++)
        ret += base64_chars[char_array_4[i]];
      i = 0;
    }
  }

  if (i)
  {
    for(j = i; j < 3; j++)
      char_array_3[j] = '\0';

    char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
    char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
    char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
    char_array_4[3] = char_array_3[2] & 0x3f;

    for (j = 0; (j < i + 1); j++)
      ret += base64_chars[char_array_4[j]];

    while((i++ < 3))
      ret += '=';

  }

  return ret;

}
//------------------------------------------------------------------------------
std::string get_boundary(const char *buffer, const int buffer_size)
{
    std::string res;
    if (!buffer_size)
    {
        res += 'A';
    }
    else
    {
        char c = buffer[0] + 1;
        if ('A' <= c && c <= 'Z') res += c;
        else res += 'A';
        
        for (int i = 0; i < buffer_size; i++)
        {
            if (res.size() > buffer_size - i) break;

            bool good = false;
            for (int j = 0; j < res.size(); j++)
            {
                if (res.at(j) != buffer[i + j])
                {
                    good = true;
                    break;
                }
            }
            if (!good)
            {
                if (res.size() == buffer_size - i)
                {
                    res += 'A';
                }
                else
                {
                    char c = buffer[i + res.size()] + 1;
                    if ('A' <= c && c <= 'Z') res += c;
                    else res += 'A';
                }
            }
        }
    }
    return res;
}
//------------------------------------------------------------------------------
bool read_post_header
(
    char *package, const int package_length,
    char ** const data, int * const data_length, std::string * const tail,
    int * const pos
)
{
/*
POST http://<host or address>:<port>/ HTTP/1.1
Host: <host or address>:<port>
Proxy-Authorization: Basic <proxy-authorization data>
Connection: Close
Content-Type: multipart/form-data; boundary=<boundary>
Content-Length: <content length>

--------------------------------------------------------------------------------
content beg
--------------------------------------------------------------------------------
--<boundary>
Content-Type: application/octet-stream
Content-Transfer-Encoding: binary

<data>--<boundary>--

--------------------------------------------------------------------------------
content end
--------------------------------------------------------------------------------
*/
    bool tail_ok = false;
    bool content_length_ok = false;
    bool content_beg_pos_ok = false;
    bool data_ok = false;

    std::string boundary_marker = "Content-Type: multipart/form-data; boundary=";
    std::string content_length_marker = "Content-Length: ";

    int content_length;

    int blank_lines_count = 0;

    int content_beg_pos;

    std::stringstream ss;
    ss.write(package, package_length);

    std::string s;

    while (true)
    {
        getline(ss, s);

        if (!strncmp(s.c_str(), boundary_marker.c_str(), boundary_marker.length()))
        {
            *tail = s.substr(boundary_marker.length(), s.length() - boundary_marker.length() - 1);
            tail_ok = true;
        }
        else if (!strncmp(s.c_str(), content_length_marker.c_str(), content_length_marker.length()))
        {
            content_length = atoi(s.substr(content_length_marker.length(), s.length() - content_length_marker.length() - 1).c_str());
            content_length_ok = true;
        }
        else if (s == "\r")
        {
            blank_lines_count++;

            if (blank_lines_count == 1)
            {
                content_beg_pos = ss.tellg();
                content_beg_pos_ok = true;
            }
            if (blank_lines_count == 2)
            {
                *data = package + (int)ss.tellg();
                *data_length = content_length - (int)ss.tellg() + content_beg_pos - tail->length() - 6;
                *tail = "--" + *tail + "--\r\n";
                *pos = package + package_length - *data - 1;
                data_ok = true;
                break;
            }
        }

        if (ss.peek() == EOF) break;
    }

    return
    tail_ok &&
    content_length_ok &&
    content_beg_pos_ok &&
    data_ok &&
    !ss.fail();
}

